from typing import Literal, Union
import numpy as np
from alitra.frame_dataclasses import Euler, PointList, Point, Transform, Translation


class FrameTransform:
    """Let (from_) be the fixed local coordinate frame that the robot operate in, and (to_) be
     in the asset-fixed global coordinate system. Further, let the relationship between the two
     reference systems be described by : p_to = rot_transform(euler_angles)*p_from + translation,
      using the zyx Euler angle rotation convention. This object allows for transform between the frames
      (from_) and (to_). Can be rewritten to accommodate for more frames.
    :param euler: Euler angles transform (zyx convention).
    :param translation: translation from frame (from_) to frame (to_), expressed in the (to_) frame.
    """

    def __init__(
        self,
        euler: Euler,
        translation: Translation,
        from_: Literal["robot", "asset"],
        to_: Literal["robot", "asset"],
    ):
        try:
            self.transform = Transform(
                translation=translation, euler=euler, from_=from_, to_=to_
            )
        except ValueError as e:
            raise ValueError(e)

    def transform_point(
        self,
        coordinates: Union[Point, PointList],
        from_: Literal["robot", "asset"],
        to_: Literal["robot", "asset"],
    ) -> Union[Point, PointList]:
        """
        Transforms a point or list of points from _from to _to (rotation and translation)
        :param coordinates: Point or PointList of coordinates to transform.
        :param from_: Source coordinate system. Must be "robot" or "asset" and different to "to_".
        :param to_: Destination coordinate system. Must be "robot" or "asset" and different to "from_".
        :return: Point or PointList with coordinates in the to_ coordinate system.
        """
        if coordinates.frame != from_:
            raise ValueError(
                f"Expected coordinates in frame {from_} "
                + f", got coordinates in frame {coordinates.frame}"
            )

        result: np.ndarray
        if from_ == self.transform.to_ and to_ == self.transform.from_:
            """Using the inverse transform"""
            result = self.transform.rotation_object.apply(
                coordinates.as_np_array() - self.transform.translation.as_np_array(),
                inverse=True,
            )

        elif from_ == self.transform.from_ and to_ == self.transform.to_:
            result = (
                self.transform.rotation_object.apply(coordinates.as_np_array())
                + self.transform.translation.as_np_array()
            )
        else:
            raise ValueError("Transform not specified")

        if isinstance(coordinates, Point):
            return Point.from_array(result, to_)
        elif isinstance(coordinates, PointList):
            return PointList.from_array(result, to_)
        else:
            raise ValueError("Incorrect input format. Must be Point or PointList.")



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Literal

import numpy as np
from scipy.spatial.transform import Rotation


@dataclass
class Transform:
    """A transform object that describe the transformation between two frames.
    Translations must be expressed in the (to_) frame"""

    translation: Translation
    euler: Euler
    from_: Literal["robot", "asset"]
    to_: Literal["asset", "robot"]
    rotation_object: Rotation = Rotation.from_euler("zyx", [0, 0, 0])

    def __post_init__(self):
        if not self.translation.from_ == self.euler.from_ == self.from_:
            raise ValueError(
                f"The from_ frames are not equal: ({self.euler.from_}), ({self.from_}), ({self.translation.from_})"
            )
        if not self.translation.to_ == self.euler.to_ == self.to_:
            raise ValueError(
                f"The to_ frames are not equal: ({self.euler.to_}),({self.to_}),({self.translation.to_})"
            )
        if not self.to_ == self.translation.frame:
            raise ValueError(
                f"The to_ frame ({self.to_}) does not match"
                + f"the translation frame ({self.translation.frame})"
            )
        self.rotation_object = Rotation.from_euler(
            "zyx", self.euler.as_np_array(), degrees=False
        )


@dataclass
class Point:
    x: float
    y: float
    frame: Literal["robot", "asset"]
    z: float = 0

    def as_np_array(self) -> np.ndarray:
        return np.array([self.x, self.y, self.z], dtype=float)

    @staticmethod
    def from_array(coordinate: np.ndarray, frame: Literal["robot", "asset"]) -> Point:
        if coordinate.shape != (3,):
            raise ValueError("Point should have shape (3,)")
        return Point(x=coordinate[0], y=coordinate[1], z=coordinate[2], frame=frame)


@dataclass
class PointList:
    points: List[Point]
    frame: Literal["robot", "asset"]

    def as_np_array(self) -> np.ndarray:
        points = []
        for point in self.points:
            points.append([point.x, point.y, point.z])
        return np.array(points, dtype=float)

    @staticmethod
    def from_array(
        point_array: np.ndarray, frame: Literal["robot", "asset"]
    ) -> PointList:
        if point_array.shape[1] != 3:
            raise ValueError("Coordinate_list should have shape (3,N)")
        points: List[Point] = []
        for point in point_array:
            points.append(Point(x=point[0], y=point[1], z=point[2], frame=frame))
        return PointList(points=points, frame=frame)


@dataclass
class Euler:
    """Euler angles where (psi,phi,theta) is rotation about the z-,y-, and x- axis respectively. as_array and from_array
    are both ordered by rotation about z,y,x. That is [psi,phi,theta]"""

    from_: Literal["robot", "asset"]
    to_: Literal["robot", "asset"]
    psi: float = 0
    phi: float = 0
    theta: float = 0

    def as_np_array(self) -> np.ndarray:
        return np.array([self.psi, self.phi, self.theta], dtype=float)

    @staticmethod
    def from_array(
        rotations: np.ndarray,
        from_: Literal["robot", "asset"],
        to_: Literal["robot", "asset"],
    ) -> Euler:
        if rotations.shape != (3,):
            raise ValueError("Coordinate_list should have shape (3,)")
        return Euler(
            psi=rotations[0], phi=rotations[1], theta=rotations[2], from_=from_, to_=to_
        )


@dataclass
class Translation:
    """Translations should be expressed in the to_ frame, which are typically the asset frame"""

    x: float
    y: float
    from_: Literal["robot", "asset"]
    to_: Literal["robot", "asset"]
    frame: Literal["asset", "robot"] = "asset"
    z: float = 0

    def as_np_array(self) -> np.ndarray:
        return np.array([self.x, self.y, self.z], dtype=float)

    @staticmethod
    def from_array(
        coordinate: np.ndarray,
        from_: Literal["robot", "asset"],
        to_: Literal["robot", "asset"],
    ) -> Translation:
        if coordinate.shape != (3,):
            raise ValueError("Point should have shape (3,)")
        return Translation(
            x=coordinate[0], y=coordinate[1], z=coordinate[2], from_=from_, to_=to_
        )
